package pongai;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;



/**************************************************************
  Tda pro obecnou neuronovou s                             
 *  @author Ondra ika
**************************************************************/
public class cNeuralNet{
	
	ArrayList aNeurons;
	ArrayList aoInputNeurons;
	ArrayList aoOutputNeurons;
	double dMinWeight;
	double dMaxWeight;
	ArrayList aoNeuronsOrder;                         // Poad vpotu neuron 
	Boolean bContainsCycles;
	
  /** Vytvo neuronovou s s danm potem (nepropojench) neuron.
   * @param iNumberOfNeurons Poten poet neuron v sti 
   * @param dMinWeight Minimln vha synapse pi generovn nhodnch vah
   * @param dMaxWeight Maximln vha synapse pi generovn nhodnch vah
   */
	cNeuralNet(int iNumberOfNeurons, double dMinWeight, double dMaxWeight){
		this.CreateNeurons(iNumberOfNeurons);
		this.aoInputNeurons  = new ArrayList();
		this.aoOutputNeurons = new ArrayList();
		this.dMinWeight = dMinWeight;
		this.dMaxWeight = dMaxWeight;
		this.aoNeuronsOrder  = null;                         // Poad vpotu neuron 
		this.bContainsCycles = null;                         // Jestli obsahuje cykly  
	}
  /** Defaultn hodnoty pro dMinWeight: 0.05, dMaxWeight: 0.30 */
	cNeuralNet(int iNumberOfNeurons){ this(iNumberOfNeurons, 0.05, 0.30); }

  /** Vytvo a vrt (nepropojen) neuron, s danm ID, s danou vnitnn funkc
   * @param iID     Pokud je 0, nov neuron bude mt vygenerovan ID uniktnm pro tuto s.
   *                Jinak pijme zadan (uniktnost se nekontroluje).
   * @param dValue  Hodnota... k niemu to nen.
   * @param dBias   Vha u x0 - pdavnho vstupu. Pouvno pi nahrvn st.
   * @param sFuncID  String ID vybran funkce. Fungovalo v JavaScriptu, v Java verzi jen Sigmoida.
   */
	public cNeuron CreateNeuron(int iID, double dValue, double dBias, String sFuncID){
		return new cNeuron(iID, dValue, dBias /*, sFuncID*/);
	}
  /** Vytvo a vrt neuron s vygenerovanm ID uniktnm pro tuto s defaultnmi i nhodnmi hodnotami */
	public cNeuron CreateNeuron(){ return new cNeuron(); }

  /** Vygeneruje dan poet neuron a vlo je do st. */
	public void CreateNeurons(int iNumberOfNeurons){
		this.aNeurons = new ArrayList();    // neurons    
		for(int i = iNumberOfNeurons; i > 0; i--){
			//this.aNeurons.push(new cNeuron());
			this.aNeurons.add(this.CreateNeuron());
		}
	}


	/** Spoji dva neurony smerovanym spojenim - u jednoho prida transmitter, u druheho receptor
   *  @param oNeuronFrom  Neuron, ze kterho spojit.
   *  @param oNeuronTo    Neuron, do kterho spojit.
   *  @param dWeight      Vha, kter se na synapsi nastav.
   */
	public boolean ConnectFromTo(cNeuron oNeuronFrom, cNeuron oNeuronTo, double dWeight){
		if( null == oNeuronFrom || null == oNeuronTo || oNeuronFrom == oNeuronTo ) return false;
		Debug.Write(Debug.E_NOTICE, 1002, "ConnectFromTo( "+(oNeuronFrom.id)+" -> "+(oNeuronTo.id)+" , W: "+dWeight+" )");

		this.aoNeuronsOrder  = null;   // Zrume ppadn spotan poad vpotu neuron 
		this.bContainsCycles = null;   // Zrume ppadn spotanou informaci, jestli obsahuje cykly 

		cSynapse oSynapse = new cSynapse(oNeuronFrom, oNeuronTo, dWeight);
		oNeuronFrom.AddTrans(oSynapse);
		oNeuronTo.AddRecep(oSynapse);
		return true;
	}
  /** Verze s nhodn nastavenou vhou. */
	public boolean ConnectFromTo(cNeuron oNeuronFrom, cNeuron oNeuronTo){
		// Nahodn vha 
		double dWeight = ( Math.random() * (this.dMaxWeight - this.dMinWeight) + this.dMinWeight )  *  ((Math.random() < 0.5) ? -1 : 1);
		return ConnectFromTo(oNeuronFrom, oNeuronTo, dWeight);
	}
	




	/**************************************************************
		Vpoet neuronov st                                       
	**************************************************************/
	public void Compute(){
		ArrayList aoCurNeurons;
		HashSet setNextNeurons = new HashSet();
		HashSet setDoneNeurons = new HashSet();
		cNeuron oCurNeuron;

		int _iRounds = 0;

		aoCurNeurons = this.aoInputNeurons;

		// Dokud jsou ve front na vpoet njak neurony... 
		while(aoCurNeurons.size() > 0) {

			Debug.Write(Debug.E_NOTICE, 1001, "Starting computing round #"+(_iRounds++)+""); ///

			// Pro vechny neurony ve front na vpoet v tomto kole... 
			Iterator i = aoCurNeurons.iterator();
			//for( var i in aoCurNeurons ){
			while(i.hasNext()){

				//oCurNeuron = aoCurNeurons[i];
				oCurNeuron = (cNeuron)i.next();
					//Debug.Write("Computing neuron { "+(oCurNeuron.id)+" }"); ///
				oCurNeuron.Compute();
				setDoneNeurons.add(oCurNeuron);


				//Debug.Write("- Following: [ "+(oCurNeuron.aTrans)+" ]"); ///

				// Pro vechny navazujc neurony... 
				Iterator j = oCurNeuron.aTrans.iterator();
				//for( var j in oCurNeuron.aTrans ){
				while(j.hasNext()){
					// var oToNeuron = oCurNeuron.aTrans[j].oTo;
					cNeuron oToNeuron = (cNeuron)( ((cSynapse)j.next()).oTo );
					// ...pokud u navazujc neuron nen vypoten, 
					if( oToNeuron != null  &&  !setDoneNeurons.contains(oToNeuron) )
						// pidme ho do fronty na vpoet v dalm kole. 
						setNextNeurons.add(oToNeuron);
				}
			}
			//aoCurNeurons = setNextNeurons.getMembers();
			//Object[] aSetMembers = setNextNeurons.toArray();
			//aoCurNeurons = new ArrayList(aSetMembers);
			aoCurNeurons = new ArrayList(setNextNeurons);
			setNextNeurons.clear();
		}

		//Debug.Write("."); ///
		Debug.Write(Debug.E_NOTICE, 1001, "Computing finished."); ///

	}


 




	/***************************************************************************************
		Rozhodnem, jestli s obsahuje smyky.                                                
		Pokud ne, vyhled bezproblmov poad vpotu neuron.                               
			Koncept prochzen st: 
			- Jsou dv mnoiny:              (lib.cSet.js) 
				- setDoneNeurons - uchovv neurony se spotenm errorem 
				- setNextNeurons - obsahuje kandidty na vpoet v dalm kole (while)
			- V kadm kole se projdou kandidti a zjist se, jestli vechny jejich 
				odchoz synapse vedou do ji spotench neuron (setDoneNeurons.has(...)).
				- Pokud ne, neuron se ponech mezi kandidty.
				- Pokud ano, neuron se spote (dError),
					odstran se z kandidt (setNextNeurons.remove(...)),
					vlo se do spotench (setDoneNeurons.put()),
					a neurony do nj vchzejc se pidaj do kandidt (setNextNeurons.put()).
			- Tento algoritmus nefunguje, pokud s obsahuje smyky.
	***************************************************************************************/
	public ArrayList FindComputingOrder(){

		// Pracovn prostor... 
		ArrayList aoCurNeurons;
		ArrayList aoNeuronsOrder = new ArrayList();
		HashSet setNextNeurons = new HashSet();
		HashSet setDoneNeurons = new HashSet();
		cNeuron oCurNeuron;
		int _iRounds = 0;

		aoCurNeurons = this.aoOutputNeurons;

		// Dokud jsou ve front na vpoet njak neurony... 
		while(aoCurNeurons.size() > 0) {
			Debug.Write(Debug.E_NOTHING, 2311, "."); ///
			Debug.Write(Debug.E_NOTICE, 2312, "Starting backprop round #"+(_iRounds++)+""); ///

			try{
			// Pro vechny neurony ve front na vpoet v tomto kole... 
			Iterator i = aoCurNeurons.iterator();
			for_aoCurNeurons:
			//for( var i in aoCurNeurons ){
			while(i.hasNext()){
				//oCurNeuron = aoCurNeurons[i];
				oCurNeuron = (cNeuron)i.next();
					Debug.Write(Debug.E_NOTICE, 2313, "- Traversing "+(oCurNeuron.toStringFull())+""); ///
				Iterator j; // for cycles 


				// Projedem navazujc neurony 
				boolean bRefersToNonProcessedNeuron = false;
				j = oCurNeuron.aTrans.iterator();
				//for( j in oCurNeuron.aTrans ){
				while(j.hasNext()){
					//var oSynapseOut = oCurNeuron.aTrans[j];
					cSynapse oSynapseOut = (cSynapse)j.next();
					if( oSynapseOut.oTo == null )  continue;  // Pseudoneuron (vstupn nebo tak nco)
					// Pokud neuron jet nem spoten dError, peskome ho, a nechme ho jet ve front na zpracovn. 
					if( !setDoneNeurons.contains(oSynapseOut.oTo) ){
						Debug.Write(Debug.E_NOTICE, 2313, "-- Skipping - has synapse to non-ready neuron."); ///
						break for_aoCurNeurons;
					}
				}

				setDoneNeurons.add(oCurNeuron);     // Pidme zpracovvan neuron mezi hotov. 
				setNextNeurons.remove(oCurNeuron);  // Odstranme zpracovvan neuron z fronty na zpracovn.
				// Neurony bez vstup peskome - to jsou vstupn pseudoneurony s hodnotami 
				if( oCurNeuron.aRecep.size() != 0 ){
					aoNeuronsOrder.add(oCurNeuron);    // Ulome zpracovvan neuron do poad.   
				}



					//Debug.Write("- Following: [ "+(oCurNeuron.aRecep)+" ]"); ///

				// Pro vechny pedchzejc neurony... 
				j = oCurNeuron.aRecep.iterator();
				//for( j in oCurNeuron.aRecep ){
				while(j.hasNext()){
					// var oFromNeuron = oCurNeuron.aRecep[j].oFrom;
					cNeuron oFromNeuron = (cNeuron)j.next();
					if( oFromNeuron == null )  continue;

					// Pokud u nen hotov, 
					if( !setDoneNeurons.contains(oFromNeuron) ){						
						// pidme ho do fronty na vpoet v dalm kole. 
						setNextNeurons.add(oFromNeuron);
					}
				}
				Debug.Write(Debug.E_NOTICE, 2313, "+ Traversed "+(oCurNeuron.toStringFull())+""); ///

			}// for( var i in aoCurNeurons )			
			}catch(Exception e){ Debug.Write(Debug.E_ERROR, 2314, "... Exception " + e); }

			// V setNextNeurons mme mnoinu neuron na dal kolo backpropagation. 
			//aoCurNeurons = setNextNeurons.getMembers();
			aoCurNeurons = new ArrayList(setNextNeurons);
			//setNextNeurons.clear();     // Neurony se odstrauj, kdy jsou spoten. 

		}// while()

		Debug.Write(Debug.E_NOTHING, 2313, "."); ///
		Debug.Write(Debug.E_NOTHING, 2313, "Traversing finished."); ///

		return this.aoNeuronsOrder = aoNeuronsOrder;
	}


	
  

	/** Serializace. Ulo neuronovou s do textu, z nho se d zrekonstruovat.
   * @see #Unserialize(String)
   */
	public String Serialize(){
		String sSaveString = "";

		sSaveString += "@neurons\n";
		sSaveString += "# id : dValue : dBias : fFunc [ : oNeuronTo.id / dWeight ]*\n";

		//for( var i in this.aoNeuronsOrder ){
		Iterator i = aoNeuronsOrder.iterator();
		while(i.hasNext()) {
			cNeuron oN = (cNeuron)i.next();
			// id, dValue, dBias, fFunc, 
			sSaveString += oN.id+":"+oN.dValue+":"+oN.dBias+":"+oN.fFunc.getID()+"";
			//for( var j in oN.aTrans ){
			Iterator j = oN.aTrans.iterator();
			while(j.hasNext()){
				// dWeight, oFrom, oTo
				cSynapse oS = (cSynapse)j.next();
				if( null == oS.oTo ) continue; // Njak pseudosynapse - asi vstupn vrstva 
				sSaveString += ":"+(oS.oTo == null ? "null" : oS.oTo.id)+"/"+oS.dWeight+"";
			}
			sSaveString += "\n";
		}

		sSaveString += "@input";
		//for( var i in this.aoInputNeurons ){ sSaveString += " "+this.aoInputNeurons[i].id; }
		i = aoInputNeurons.iterator();
		while(i.hasNext()) { cNeuron oN = (cNeuron)i.next(); sSaveString += " "+oN.id; }
		sSaveString += "\n";

		sSaveString += "@output";
		//for( var i in this.aoOutputNeurons ){ sSaveString += " "+this.aoOutputNeurons[i].id; }
		i = aoOutputNeurons.iterator();
		while(i.hasNext()) { cNeuron oN = (cNeuron)i.next(); sSaveString += " "+oN.id; }
		sSaveString += "\n";

		sSaveString += "@cycles "+this.bContainsCycles+"\n";

		return sSaveString;
	}

  
  
  
  
  
  
  
  /** Deserializace. Ze Stringu zrekonstruuje v nm zapsanou s. */
	public void Unserialize(String sSaveString) throws Exception {
		String[] asLines = sSaveString.split("\n", -1);

		// Vsledky 
		ArrayList aoNeurons = new ArrayList();
		ArrayList aoInputNeurons = new ArrayList();
		ArrayList aoOutputNeurons = new ArrayList();
		HashMap aoPendingSynapses = new HashMap();
		String sNetType = null;

		// 0 - nothing; 1 - neurons 
		int iMode = 0;

		// Pro vechny dky 
		//for( var i in asLines ){
		for(String sLine : asLines){
			if('#' == sLine.charAt(0)) continue;
			
			//String[] asReResult = sLine.match( /^@(\w+)(\s+(.*))?/ );
			Pattern p = Pattern.compile("^@(\\w+)(\\s+(.*))?");
			Matcher m = p.matcher(sLine);
			//if("@" == sLine.charAt(0)){
			//if(null != asReResult){
			if(m.matches()){
				String sKeywordParams = m.group(3);
				iMode = 0;
				ArrayList aoIONeurons = aoInputNeurons;
				//if("neurons" == sLine.substring()){
				String sKeyword = m.group(1);
				if(sKeyword == "neurons") iMode = 1;
				if(sKeyword == "nettype")	sNetType = sKeywordParams;
				if(sKeyword == "output")                   // Abych nemusel pst znova ten samej kd, udlme fintu. 
					aoIONeurons = aoOutputNeurons;           // aoIONeurons je bu? aoO... nebo aoI..., podle keywordu.
				if(sKeyword == "output" || sKeyword == "input"){
					String saNeuronIDs = sKeywordParams;
					String[] asNeuronIDs = saNeuronIDs.split(" ");
					for( String sInpNeuID : asNeuronIDs ){
						int iInpNeuID = Integer.valueOf(sInpNeuID);
						if( null == aoNeurons.get(iInpNeuID) )
							throw new Exception("Undefined neuron: ID = "+iInpNeuID+"");
						aoIONeurons.add( aoNeurons.get(iInpNeuID) );
					}
				}
				
				if(sKeyword == "cycles"){
					if     (sKeywordParams == "false") this.bContainsCycles = false;
					else if(sKeywordParams == "true")  this.bContainsCycles = true;
					else                               this.bContainsCycles = null;
				}
				
			}// if(m.matches()){
			
			// Na zatku nen "@" 
			else if(iMode == 1){
				// id : dValue : dBias : fFunc [ : oNeuronTo.id / dWeight ]* 
				String[] asParts = sLine.split(":");
				if( asParts.length < 4 )
					throw new Exception("cNeuralNet::Unserialize(): Neuron has to be defined as \"id : dValue : dBias : fFunc [ : oNeuronTo.id / dWeight ]*\"");
															 // iID, dValue, dBias, sFuncID 
				cNeuron oNeuron = new cNeuron(Integer.valueOf(asParts[0]), Double.valueOf(asParts[1]), Double.valueOf(asParts[2])/*, (asParts[3])*/);
				aoNeurons.ensureCapacity(oNeuron.id + 1);
				aoNeurons.set(oNeuron.id, oNeuron);

				for( int iTrans = 4; iTrans < asParts.length; iTrans++ ){
					// oNeuronTo.id / dWeight
					String[] asSynapse = asParts[iTrans].split("/");
					int iToId = Integer.valueOf(asSynapse[0]);
					double dWeight = Double.valueOf(asSynapse[1]);

					// Pokud nen navazujc neuron jet vytvoen, dme si spojen do fronty pro neuron s danm ID. 
					if( null == aoNeurons.get(iToId) ){
						//throw new Exception("No such neuron ID");
						if( null == aoPendingSynapses.get(iToId) ) aoPendingSynapses.put(iToId, new ArrayList());
						((ArrayList)(aoPendingSynapses.get(iToId))).add( new cSynapse( oNeuron, null, dWeight) );
					}
					// Jinak tu synapsi vytvome.
					else
						this.ConnectFromTo(oNeuron, (cNeuron)aoNeurons.get(iToId), dWeight);

					// Pokud na tento neuron ji ekaj njak pchoz synapse, vytvome.
					ArrayList aoSynapses;
					if( null != (aoSynapses = (ArrayList)aoPendingSynapses.get(oNeuron.id)) /* && "object" == typeof aoPendingSynapses[oNeuron.id] */ ){
						Iterator iSynIndex = aoSynapses.iterator();
						//for( var iSynIndex in aoPendingSynapses.get(oNeuron.id) )
						while(iSynIndex.hasNext()){
							cSynapse oSynapse = (cSynapse)iSynIndex.next();
							this.ConnectFromTo(
								oSynapse.oFrom,
								oNeuron,
								oSynapse.dWeight );
						}
					}
				}
			}// Na zatku nen "@" - else if(iMode == 1)

		}// Pro vechny dky - for( var i in asLines )

		this.aoNeuronsOrder = aoNeurons;

	}



}// class cNeuralNet









