/*
 * cNeuralNetTeacher.java
 *
 * Created on 1. leden 2006, 11:37
 */

package pongai;

import java.util.Date;
import java.util.HashSet;
import java.util.ArrayList;



/**************************************************************
    class cNeuralNetTeacher                                    
	Implementation of supervised learning with backpropagation.   
 *  @author Ondra ika
**************************************************************/
public class cNeuralNetTeacher {
	
	cNeuralNetIOAdapter oNetAdapter;
	cNeuralTrainSet     oTrainSet;
	cNeuralTrainSetItem oCurTrainSetItem;

	double  dLearn;                    // Learning rate - correction amount (LAMBDA) 
	double  dMomentum;                // Momentum - net's momentum (setrvanost)    
	int     iMaxRepeats;                  // Maximum mumber of repeats of training step 
	double  dTargetError;              // Target error - stop training when reached ?


	int iTurns;                       // Number of network computations        
	int  iCorrections;                // Number of backpropagation corrections 
	double dLastErrorLevel;           // Average error on output neurons (error sum / # of output neurons)
	double dLastErrorSum;             // Sum of errors on all output neurons   
	
	class cLearnRateInterval {
		public double dError;
		public double dLearnRate;
		public cLearnRateInterval(double dError, double dLearnRate){
			this.dError = dError;
			this.dLearnRate = dLearnRate;
		}
	}
	cLearnRateInterval[] aodLearnRateIntervals;
	
	// adErrorLevels = [1.0];
	// oAvgErrorLevels = new cAverage();  oAvgErrorLevels.AddSample( adErrorLevels[0]);
	ArrayList adErrorLevels = new ArrayList();
	cAverage oAvgErrorLevels = new cAverage();

	cAverage oAvgCorrectionTime = new cAverage();
	int iLastTimeMs = 0;
	
	
	/** Vytvo uitele neuronov st.
   @param oNeuralNetIOAdapter  NN adaptr s pidruenou neuronovou st.
   @param oTrainSet            Trnovac mnoina pro tohoto uitele. */
	public cNeuralNetTeacher(cNeuralNetIOAdapter oNeuralNetIOAdapter, cNeuralTrainSet oTrainSet){
		this(oNeuralNetIOAdapter, oTrainSet, 0.2, 0.5, 0.03, 1000);
	}
  /** Vytvo uitele neuronov st.
   @param oNeuralNetIOAdapter  NN adaptr s pidruenou neuronovou st.
   @param oTrainSet            Trnovac mnoina pro tohoto uitele.
   @param dLearn               Learning rate - correction amount (LAMBDA) 
   @param dMomentum            Momentum - net's momentum (setrvanost) (does not work)
   @param dTargetError         Target error - stop training when reached?
   @param iMaxRepeats          Maximum mumber of repeats of training step.
  */
	public cNeuralNetTeacher(cNeuralNetIOAdapter oNeuralNetIOAdapter, cNeuralTrainSet oTrainSet,
				double dLearn, double dMomentum, double dTargetError, int iMaxRepeats){

		this.oNetAdapter = oNeuralNetIOAdapter;
		//this.oTrainSet   = oTrainSet;
    this.SetTrainSet(oTrainSet);
		this.oCurTrainSetItem = null;

		this.dLearn = dLearn;                 // Learning rate - correction amount (LAMBDA) 
		this.dMomentum = dMomentum;           // Momentum - net's momentum (setrvanost)    		
		this.dTargetError = dTargetError;     // Target error - stop training when reached ?
		this.iMaxRepeats  = iMaxRepeats;      // Maximum mumber of repeats of training step 


		this.iTurns = 0;                      // Number of network computations        
		this.iCorrections = 0;                // Number of backpropagation corrections 
		this.dLastErrorSum = 0.0;             // Sum of errors on all output neurons   
		this.dLastErrorLevel = 0.0;           // Average error on output neurons (error sum / # of output neurons)
		//this.adErrorLevels = [1.0];
		//this.oAvgErrorLevels = new cAverage(); this.oAvgErrorLevels.AddSample(this.adErrorLevels[0]);
		this.adErrorLevels = new ArrayList();
		this.oAvgErrorLevels = new cAverage();
		
		this.aodLearnRateIntervals = null;

		this.oAvgCorrectionTime = new cAverage();
		this.iLastTimeMs = 0;
	}

  /** Vrt pidruenou s pidruenho NN adaptru. */
	public cNeuralNet GetNet()   { return this.oNetAdapter.oNeuralNet; }
  /** Vrt posledn souet chyb. */
	public double GetErrorSum()  { return this.dLastErrorSum; }
  /** Vrt posledn rove chyby. */
	public double GetErrorLevel(){ return this.dLastErrorLevel; }

  /** Vrt poet cykl vpotu (vybrn vstupu a vpoet). */
	public int GetTurnsCount(){ return this.iTurns; }
  /** Vrt poet oprav. */
	public int GetCorrectionsCount(){ return this.iCorrections; }
  /** Vynuluje potadla cyklu uen a cyklu oprav. */
	public void ResetCounters(){ this.iTurns  = 0; this.iCorrections = 0; }
  
  /** Piad novou trnovac mnoinu. */
  public void SetTrainSet(cNeuralTrainSet oTrainSet){ this.oTrainSet = oTrainSet; }

	/** Jedno kolo vpotu - vybrn vstupu a vpoet.  */
	public void Turn(){
		this.iTurns++;
		this.oCurTrainSetItem = this.oTrainSet.GetRandom();
    if(null == this.oCurTrainSetItem){
      Debug.Write(Debug.E_ERROR, "cNeuralNetTeacher::Turn(): this.oCurTrainSetItem == null!!");
      return;
    }
		//this.oCurTrainSetItem = this.oTrainSet.GetAt(2);
		Debug.Write(Debug.E_NOTICE, 1006, "Vstupn hodnoty: [" + this.oCurTrainSetItem.adI + "]"); ///
		Debug.Write(Debug.E_NOTICE, 1006, "Chtn  hodnoty: [" + this.oCurTrainSetItem.adO + "]"); ///

		this.oNetAdapter.SetInputValues(this.oCurTrainSetItem.adI);
		this.oNetAdapter.oNeuralNet.Compute();
	}
	

	
	/*********************************************************************************
	  cNeuralNetTeacher::PerformCorrection()                                          
		Pizpsoben st: Porovnn rozdlu a oprava vah. 
		
		Koncept prochzen st:
		
		- Jsou dv mnoiny:              (lib.cSet.js)
		  - setDoneNeurons - uchovv neurony se spotenm errorem 
		  - setNextNeurons - obsahuje kandidty na vpoet v dalm kole (while)
		- V kadm kole se projdou kandidti a zjist se, jestli vechny jejich 
		  odchoz synapse vedou do ji spotench neuron (setDoneNeurons.has(...)).
		  - Pokud ne, neuron se ponech mezi kandidty.
		  - Pokud ano, neuron se spote (dError),
		    odstran se z kandidt (setNextNeurons.remove(...)),
		    vlo se do spotench (setDoneNeurons.put()),
		    a neurony do nj vchzejc se pidaj do kandidt (setNextNeurons.put()).
		- Tento algoritmus nefunguje, pokud s obsahuje smyky.
		- Plnuju algoritmus dt pmo do cNeuralNetwork, kde by ml za kol rozhodnout
		  o tom, jestli s obsahuje smyky, a pokud ne, sestavit poad vpotu.
		
	*********************************************************************************/
	public boolean PerformCorrection(){
		this.iCorrections++;
		//var _dw = null;
		
		Date oTimeStart = new Date();
		double dErrorSum = 0.0;	
		cNeuralNet oNeuralNet = this.oNetAdapter.oNeuralNet;
		
		// Zskn hodnot 
		double[] adVystupyReal = this.oNetAdapter.GetOutputValues();
		double[] adVystupyWant = this.oCurTrainSetItem.adO;
		if(adVystupyWant.length < adVystupyReal.length)
			return false;
			
		//String s = ArrayMap( adVystupyReal, function(d){ return cRound.Round(d,4); } );
		Debug.Write("Vstupn hodnoty: [" + cRound.implode(adVystupyReal) + "]"); ///

		
		/*for( var i in adVystupyReal ){
			//dErrorSum += adVystupyDiff[i] = adVystupyWant[i] - adVystupyReal[i];
			//adWeightDeltas[i] = this.oNetAdapter.oNeuralNet.fFunc( adVystupyReal[i] ) * adVystupyDiff[i];
			// error = OutReal * (1  OutReal) * (OutWant  OutReal)
		}/**/
			

		// Pracovn prostor... 
		ArrayList<cNeuron> aoCurNeurons;
		HashSet setNextNeurons = new HashSet();
		HashSet setDoneNeurons = new HashSet();
		//cNeuron oCurNeuron;
		int _iRounds = 0;
		
		
		/* ----------------- Vstupn vrstva ----------------- */
		
		Debug.Write(Debug.E_NOTICE, 2331, "."); ///
		Debug.Write(Debug.E_NOTICE, 2331, "Starting backprop round # -1  - output layer"); ///
			
		// Pro vechny vstupn neurony st - nastavme jim Error. 
		int i = 0;
		for( Object o : oNeuralNet.aoOutputNeurons ){
			cNeuron oCurNeuron  = (cNeuron)o;
			  Debug.Write(Debug.E_NOTICE, 2332, "- Correcting "+(oCurNeuron.toStringFull())+""); ///
			setDoneNeurons.add(oCurNeuron);
			
			// -- http://www.tek271.com/articles/neuralNet/IntoToNeuralNets.html -- 
			        /* derivace Sigmoidy */          /* dDiff */
			// e =  (   z * (1  z)       )     *     ( y  z );    z == dOutReal,  y == dOutWant 
			
			// Rozdl mezi vstupem a vstupem 
			double dOutReal = oCurNeuron.dValue;
			double dDiff = adVystupyWant[i] - dOutReal;
			// Chyba tohoto vst. neur.
			oCurNeuron.dError = oCurNeuron.fFunc.derivation(dOutReal) * dDiff;
			  Debug.Write(Debug.E_NOTICE, 2333, "("+oCurNeuron.id+") diff: "+cRound.Round(dDiff, 5)+" err: "+cRound.Round(oCurNeuron.dError, 5)); ///
			dErrorSum += Math.abs(oCurNeuron.dError);

			// Dq = e * lambda;
			double dPom = oCurNeuron.dError * this.dLearn;
			  Debug.Write(Debug.E_NOTICE, 2334, "("+oCurNeuron.id+") err * lambda: " +cRound.Round(dPom, 5)); ///
				
			oCurNeuron.dBias += dPom;
			
			for( Object oSyn : oCurNeuron.aRecep ){
				cSynapse oSynapseIn = (cSynapse)oSyn;
				// Dw[i] = Dq * vstup_neuronu[i]; 
				oSynapseIn.dWeight += dPom * oSynapseIn.GetInput(); /// Dno do neuronu 
				setNextNeurons.add(oSynapseIn.oFrom);
			}
			Debug.Write(Debug.E_NOTICE, 2335, "+ Corrected "+(oCurNeuron.toStringShort())+""); ///
			i++;
		}// for( Object o : oNeuralNet.aoOutputNeurons )
		aoCurNeurons = new ArrayList(setNextNeurons);
		//setNextNeurons.clear();     // Neurony se odstra?uj, kdy jsou spoten. 



		// --- Zpracovn error sum --- //
		this.dLastErrorSum = dErrorSum;
		this.dLastErrorLevel = dErrorSum / oNeuralNet.aoOutputNeurons.size();
  		Debug.Write(Debug.E_NOTICE, 2336, "Error Sum: " +cRound.Round(dErrorSum, 5)); ///
		this.adErrorLevels.add(this.dLastErrorLevel);
		if(this.dLastErrorLevel < this.dTargetError)
			; // Nco udlat. 
			
		
		// Pokud nen jet nalezeno poad vpotu neuron, vypotme ho te.
    if(null == this.oNetAdapter.oNeuralNet.aoNeuronsOrder)
      this.oNetAdapter.oNeuralNet.FindComputingOrder();
    
		aoCurNeurons = this.oNetAdapter.oNeuralNet.aoNeuronsOrder;
		// Neurony z vstupn vrstvy u potat nebudeme. 
		ArrayList<cNeuron> aoCurNeurons2 = new ArrayList();
		for( cNeuron oN : aoCurNeurons ){
			if( !setDoneNeurons.contains(oN) )
				aoCurNeurons2.add( oN );
		}/**/
		aoCurNeurons = aoCurNeurons2;
		
		

		/* ----------------- Zbyl vrstvy... ----------------- */

		// Dokud jsou ve front na vpoet njak neurony... 
		while(aoCurNeurons.size() > 0) {
			Debug.Write(Debug.E_NOTHING, 2321, "."); ///
			Debug.Write(Debug.E_NOTICE, 2322, "Starting backprop round #"+(_iRounds++)+""); ///

			try{
			// Pro vechny neurony ve front na vpoet v tomto kole... 
			for_aoCurNeurons:
			for( cNeuron oCurNeuron : aoCurNeurons ){
				//oCurNeuron = aoCurNeurons[i];
          Debug.Write(Debug.E_NOTICE, 2323, "- Correcting "+(oCurNeuron.toStringShort())+""); ///
					
				// Pokud u je neuron hotov, vyndme ho z fronty a peskome.
				// Tk se vstupn vrstvy - poad vpotu je vypoten pedem. 
				// if( !setDoneNeurons.has(oCurNeuron) ) continue;
				// Tak nakonec ne - vyfiltruju si to u ped cyklem.
				
				// Seteme errory z navazujcch neuron, vynsoben vahami synaps do nich  [ g = SUM( m[i] * e[i] ) ]
				dErrorSum = 0.0;
				for( Object oSynOut : oCurNeuron.aTrans ){
					cSynapse oSynapseOut = (cSynapse)oSynOut;
					dErrorSum += oSynapseOut.oTo.dError * oSynapseOut.dWeight;
				}
				
				// Uloen chyby tohoto vst. neur.
				oCurNeuron.dError = dErrorSum * oCurNeuron.fFunc.derivation(oCurNeuron.dValue);
					//Debug.Write("("+oCurNeuron.id+")  err: "+Round(oCurNeuron.dError, 5)); ///

				double dPom  = oCurNeuron.dError * this.dLearn;

				oCurNeuron.dBias += dPom;

					//Debug.Write("- Following: [ "+(oCurNeuron.aRecep)+" ]"); ///
				// Pro vechny pedchzejc neurony... 
				for( Object oSynIn : oCurNeuron.aRecep ){
					cSynapse oSynapseIn = (cSynapse)oSynIn;
					cNeuron oFromNeuron = oSynapseIn.oFrom;
					if( oFromNeuron == null )  continue;
					// upravme vhy synaps vedoucch z nj sem, 
					oSynapseIn.dWeight += oFromNeuron.dValue * dPom;
				}
				
				Debug.Write(Debug.E_NOTICE, 2323, "+ Corrected "+(oCurNeuron.toStringShort())+""); ///
				
			}// for( var i in aoCurNeurons )			
			}catch(Exception e){ Debug.Write(Debug.E_ERROR, 2326, "... Exception " + e); }
			
			// V setNextNeurons mme mnoinu neuron na dal kolo backpropagation. 
			//aoCurNeurons = setNextNeurons.getMembers();
			break;

		}// while()
		
		
		this.iLastTimeMs = (int)( (new Date()).getTime() - oTimeStart.getTime() );
		this.oAvgCorrectionTime.AddSample( this.iLastTimeMs );
		this.oAvgErrorLevels.AddSample( this.dLastErrorLevel );
		if( this.iCorrections >= 20 )
			this.oAvgErrorLevels.RemSample(  (Double)this.adErrorLevels.get(this.adErrorLevels.size()-20) );
		
		
		
		Debug.Write(Debug.E_NOTHING, 2324, "."); ///
		Debug.Write(Debug.E_NOTICE, 2324, "Correcting finished. Time: "+this.iLastTimeMs+" ms; Overall avg time: "+this.oAvgCorrectionTime+" ms"); ///
		
		return true;

	} // cNeuralNetTeacher.prototype.PerformCorrection() 
	
	
	
	
	/** Vrac prmr z poslednch iCorrections opravovn. */
	public double GetErrorLevelsAvg(int iCorrections, int iBackOffset){
		if(iCorrections == 0){
			return this.oAvgErrorLevels.GetAverage();
		}
	
		//if(iBackOffset == 0) iBackOffset = 0;
		int iIndex1 = Math.max(0, this.adErrorLevels.size() - iCorrections - iBackOffset);
		int iIndex2 = Math.max(0, this.adErrorLevels.size() - 1 - iBackOffset);
		
		double dAvg = 0.0;
		for( int i = iIndex1; i < iIndex2; i++ ){
			dAvg += (Double)this.adErrorLevels.get(i);
		}
		return dAvg / (iIndex2 - iIndex1 + 1);
	}
	
	
	/** Sets the [error <-> learn rate ] correlation intervals */
	public void SetLearnRateIntervals(cLearnRateInterval[] aodLearnRateIntervals){
		this.aodLearnRateIntervals = aodLearnRateIntervals;
		/*this.aodLearnRateIntervals = [
			{dError: 1.0, dLearnRate: 0.7},
			{dError: 0.8, dLearnRate: 0.5},
			{dError: 0.3, dLearnRate: 0.3},
			{dError: 0.1, dLearnRate: 0.1}
		];*/
	}
	
	/** Adjusts Learn Rate (LAMBDA) according to current average error sum */
	public boolean AdjustLearnRate(){
		if(null == this.aodLearnRateIntervals) return false;
		
		double dErrorLevelAvg = this.GetErrorLevelsAvg(0, 60);
		for( int i = this.aodLearnRateIntervals.length-1;  i >= 0;  i-- ){
			if( dErrorLevelAvg <= this.aodLearnRateIntervals[i].dError ){
				this.dLearn = this.aodLearnRateIntervals[i].dLearnRate;
				break;
			}
		}
		return true;
	}
	
}// class cNeuralNetTeacher






/*
	oVzor = oTrainSet.GetRandom();
	oVzor = aoadVzory[2];
	
	_dw.Write("Vstupn hodnoty: [" + oVzor.adI + "]"); ///

	oNetAdapter.SetInputValues(oVzor.adI);
	oNetAdapter.oNeuralNet.Compute();

	var adVystupyReal = oNetAdapter.GetOutputValues();
	var dError = oNetAdapter.GetError();

	var s = ArrayMap( adVystupyReal,(d){ return RoundPrec(d, 100); } );
	_dw.Write("Vstupn hodnoty: " + s); ///
	
	for( var i = 0; i < 10; i++){
		Debug.Write("--- oNetTeacher.Turn(); ---"); ///
		oNetTeacher.Turn();
		Debug.Write("--- oNetTeacher.PerformCorrection(); ---"); ///
		oNetTeacher.PerformCorrection();
	}
*/